# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/02_timestep.ipynb (unless otherwise specified).

__all__ = ['solution', 'check_model_termination', 'reset_initial_conditions', 'update_time']

# Cell
from .solution import *
from .initialize import calculate_HI_linear, calculate_HIGC
import numpy as np
import pandas as pd

# Cell
def solution(InitCond,ClockStruct,weather_step,Outputs,Farmer):
    """
    Function to perform AquaCrop-OS solution for a single time step



    *Arguments:*\n

    `InitCond` : `InitCondClass` :  containing current model paramaters

    `ClockStruct` : `ClockStructClass` :  model time paramaters

    `weather_step`: `np.array` :  containing P,ET,Tmax,Tmin for current day

    `Outputs` : `OutputClass` :  object to store outputs

    `Farmer` : `FarmerClass` :  irrigation scheduler

    *Returns:*

    `NewCond` : `InitCondClass` :  containing updated model paramaters

    `Outputs` : `OutputClass` :  object to store outputs



    """


    # Unpack structures
    Soil = InitCond.ParamStruct.Soil
    CO2 = InitCond.ParamStruct.CO2
    if InitCond.ParamStruct.WaterTable == 1:
        Groundwater = InitCond.ParamStruct.zGW.loc[ClockStruct.StepStartTime]
    else:
        Groundwater = 0



    P = weather_step[2]
    #P = 0
    Tmax = weather_step[1]
    Tmin = weather_step[0]
    Et0 = weather_step[3]



    # Store initial conditions in structure for updating %%
    NewCond = InitCond
    # Check if growing season is active on current time step %%
    if ClockStruct.SeasonCounter >= 0:
        # Check if in growing season
        CurrentDate = ClockStruct.StepStartTime
        PlantingDate = ClockStruct.PlantingDates[ClockStruct.SeasonCounter]
        HarvestDate = ClockStruct.HarvestDates[ClockStruct.SeasonCounter]

        NewCond.CurrentDate = CurrentDate


        if (PlantingDate <= CurrentDate) and \
        (HarvestDate >= CurrentDate) and \
        (NewCond.CropMature == False) and \
        (NewCond.CropDead == False):
            GrowingSeason = True
        else:
            GrowingSeason = False

        # Assign crop, irrigation management, and field management structures
        Crop = NewCond.ParamStruct.Seasonal_Crop_List[ClockStruct.SeasonCounter]

        IrrMngt = Crop.IrrMngt



        if GrowingSeason == True:
            FieldMngt = Crop.FieldMngt
        else:
            FieldMngt = NewCond.ParamStruct.FallowFieldMngt

    else:
        # Not yet reached start of first growing season
        GrowingSeason = False
        # Assign crop, irrigation management, and field management structures
        # Assign first crop as filler crop
        Crop = NewCond.ParamStruct.Fallow_Crop
        Crop.Aer = 5; Crop.Zmin = 0.3
        IrrMngt = Crop.IrrMngt
        IrrMngt.IrrMethod =  -99
        FieldMngt = NewCond.ParamStruct.FallowFieldMngt

    # Increment time counters %%
    if GrowingSeason == True:
        # Calendar days after planting
        NewCond.DAP = NewCond.DAP+1
        # Growing degree days after planting
        GDD,NewCond = growing_degree_day(Crop,NewCond,Tmax,Tmin)
        NewCond.GDD = GDD
        NewCond.GrowingSeason = True
    else:
        NewCond.GrowingSeason = False

        # Calendar days after planting
        NewCond.DAP = 0
        # Growing degree days after planting
        GDD = 0.3
        NewCond.GDDcum = 0


    # save current timestep counter
    NewCond.TimeStepCounter = ClockStruct.TimeStepCounter

    NewCond.P = weather_step[2]
    NewCond.Tmax = weather_step[1]
    NewCond.Tmin = weather_step[0]
    NewCond.Et0 = weather_step[3]


    # Run simulations %%
    # 1. Check for groundwater table
    NewCond,Soil = check_groundwater_table(ClockStruct,Soil,NewCond)

    # 2. Root development
    NewCond = root_development(Crop,Soil,NewCond,GDD,GrowingSeason)

    # 3. Pre-irrigation
    NewCond, PreIrr = pre_irrigation(Soil,Crop,NewCond,GrowingSeason,IrrMngt)

    # 4. Drainage
    NewCond,DeepPerc,FluxOut = drainage(Soil,NewCond)

    # 5. Surface runoff
    Runoff,Infl,NewCond = rainfall_partition(P,Soil,FieldMngt,NewCond)

    # 6. Irrigation
    NewCond, Irr = irrigation(NewCond,IrrMngt,Crop,Soil,ClockStruct,GrowingSeason,P,Runoff,Farmer)


    # 7. Infiltration
    NewCond,DeepPerc,RunoffTot,Infl,FluxOut = infiltration(Soil,NewCond,Infl,Irr,IrrMngt,FieldMngt,FluxOut,DeepPerc,Runoff,GrowingSeason)

    # 8. Capillary Rise
    NewCond,CR = capillary_rise(Soil,NewCond,FluxOut)

    # 9. Check germination
    NewCond = germination(NewCond,Soil,Crop,GDD,GrowingSeason)

    # 10. Update growth stage
    NewCond = growth_stage(Crop,NewCond,GrowingSeason)

    # 11. Canopy cover development
    NewCond = canopy_cover(Crop,Soil,NewCond,GDD,Et0,GrowingSeason)

    # 12. Soil evaporation
    NewCond,Es,EsPot = soil_evaporation(ClockStruct,Soil,Crop,IrrMngt,FieldMngt,NewCond, \
                                              Et0,Infl,P,Irr,GrowingSeason)

    # 13. Crop transpiration
    Tr,TrPot_NS,TrPot,NewCond,IrrNet = transpiration(Soil,Crop,IrrMngt,NewCond,Et0,CO2, \
                                                         GrowingSeason,GDD)

    # 14. Groundwater inflow
    NewCond,GwIn = groundwater_inflow(Soil,NewCond)


    # 15. Reference harvest index
    NewCond = HIref_current_day(NewCond,Crop,GrowingSeason)

    # 16. Biomass accumulation
    NewCond = biomass_accumulation(Crop,NewCond,Tr,TrPot_NS,Et0,GrowingSeason)


    # 17. Harvest index
    NewCond = harvest_index(Soil,Crop,NewCond,Et0,Tmax,Tmin,GrowingSeason)


    # 18. Crop yield
    if GrowingSeason == True:
        # Calculate crop yield (tonne/ha)
        NewCond.Y = (NewCond.B/100)*NewCond.HIadj
        #print( ClockStruct.TimeStepCounter,(NewCond.B/100),NewCond.HIadj)
        # Check if crop has reached maturity
        if ((Crop.CalendarType == 1) and (NewCond.DAP >= Crop.Maturity)) \
        or ((Crop.CalendarType == 2) and (NewCond.GDDcum >= Crop.Maturity)):
            # Crop has reached maturity
            NewCond.CropMature = True

    elif GrowingSeason == False:
        # Crop yield is zero outside of growing season
        NewCond.Y = 0




    # 19. Root zone water
    Wr,_Dr,_TAW,_thRZ = root_zone_water(Soil,Crop,NewCond)



    # 20. Update net irrigation to add any pre irrigation
    IrrNet = IrrNet+PreIrr
    NewCond.IrrNetCum = NewCond.IrrNetCum+PreIrr


    # Update model outputs %%
    row_day = ClockStruct.TimeStepCounter
    row_gs = ClockStruct.SeasonCounter


    # Irrigation
    if GrowingSeason == True:
        if IrrMngt.IrrMethod == 4:
            # Net irrigation
            IrrDay = IrrNet
            IrrTot = NewCond.IrrNetCum
        else:
            # Irrigation
            IrrDay = Irr
            IrrTot = NewCond.IrrCum

    else:
        IrrDay = 0
        IrrTot = 0

        NewCond.Depletion =  _Dr.Rz
        NewCond.TAW = _TAW.Rz


    # Water contents
    Outputs.Water[row_day,:3] = np.array([ClockStruct.TimeStepCounter,GrowingSeason,NewCond.DAP])
    Outputs.Water[row_day,3:] = NewCond.th*NewCond.TAW

    # Water fluxes
    Outputs.Flux[row_day,:] = [ClockStruct.TimeStepCounter,\
                                          ClockStruct.SeasonCounter,NewCond.DAP,Wr,NewCond.zGW,\
                                          NewCond.SurfaceStorage,IrrDay,\
                                          Infl,Runoff,DeepPerc,CR,GwIn,Es,EsPot,Tr,P]

    # Crop growth
    Outputs.Growth[row_day,:] = [ClockStruct.TimeStepCounter,ClockStruct.SeasonCounter,NewCond.DAP,GDD,\
                                         NewCond.GDDcum,NewCond.Zroot,\
                                         NewCond.CC,NewCond.CC_NS,NewCond.B,\
                                         NewCond.B_NS,NewCond.HI,NewCond.HIadj,\
                                         NewCond.Y]

    # Final output (if at end of growing season)
    if ClockStruct.SeasonCounter > -1:
        if ((NewCond.CropMature == True) \
            or (NewCond.CropDead == True) \
            or (ClockStruct.HarvestDates[ClockStruct.SeasonCounter] == ClockStruct.StepEndTime )) \
        and (NewCond.HarvestFlag == False):

            # Store final outputs
            Outputs.Final.loc[ClockStruct.SeasonCounter] = [ClockStruct.SeasonCounter,Crop.Name,\
                                                            ClockStruct.StepEndTime,ClockStruct.TimeStepCounter,\
                                                            NewCond.Y,IrrTot]

            # Set harvest flag
            NewCond.HarvestFlag = True



    return NewCond,Outputs

# Cell
def check_model_termination(ClockStruct,InitCond):
    """
    Function to check and declare model termination


    *Arguments:*\n

    `ClockStruct` : `ClockStructClass` :  model time paramaters

    `InitCond` : `InitCondClass` :  containing current model paramaters

    *Returns:*

    `ClockStruct` : `ClockStructClass` : updated clock paramaters


    """

    ## Check if current time-step is the last
    CurrentTime = ClockStruct.StepEndTime
    if CurrentTime < ClockStruct.SimulationEndDate:
        ClockStruct.ModelTermination = False
    elif CurrentTime >= ClockStruct.SimulationEndDate:
        ClockStruct.ModelTermination = True


    ## Check if at the end of last growing season ##
    # Allow model to exit early if crop has reached maturity or died, and in
    # the last simulated growing season
    if (InitCond.HarvestFlag == True) \
    and    (ClockStruct.SeasonCounter == ClockStruct.nSeasons-1):

            ClockStruct.ModelTermination = True



    return ClockStruct

# Cell
def reset_initial_conditions(ClockStruct,InitCond):

    """
    Function to reset initial model conditions for start of growing
    season (when running model over multiple seasons)

    *Arguments:*\n

    `ClockStruct` : `ClockStructClass` :  model time paramaters

    `InitCond` : `InitCondClass` :  containing current model paramaters

    `weather_df`: `pandas.DataFrame` :  weather data for simulation period


    *Returns:*

    `InitCond` : `InitCondClass` :  containing reset model paramaters



    """

    ## Extract crop type ##
    CropType = InitCond.ParamStruct.CropChoices[ClockStruct.SeasonCounter]

    ## Extract structures for updating ##
    Soil = InitCond.ParamStruct.Soil
    Crop = InitCond.ParamStruct.Seasonal_Crop_List[ClockStruct.SeasonCounter]
    FieldMngt = Crop.FieldMngt
    CO2 = InitCond.ParamStruct.CO2
    CO2_data = InitCond.ParamStruct.CO2data

    ## Reset counters ##
    InitCond.AgeDays = 0
    InitCond.AgeDays_NS = 0
    InitCond.AerDays = 0
    InitCond.IrrCum = 0
    InitCond.DelayedGDDs = 0
    InitCond.DelayedCDs = 0
    InitCond.PctLagPhase = 0
    InitCond.tEarlySen = 0
    InitCond.GDDcum = 0
    InitCond.DaySubmerged = 0
    InitCond.IrrNetCum = 0
    InitCond.DAP = 0

    InitCond.AerDaysComp = np.zeros(int(Soil.nComp))

    ## Reset states ##
    # States
    InitCond.PreAdj = False
    InitCond.CropMature = False
    InitCond.CropDead = False
    InitCond.Germination = False
    InitCond.PrematSenes = False
    InitCond.HarvestFlag = False



    # Harvest index
    # HI
    InitCond.Stage = 1
    InitCond.Fpre = 1
    InitCond.Fpost = 1
    InitCond.fpost_dwn = 1
    InitCond.fpost_upp = 1

    InitCond.HIcor_Asum = 0
    InitCond.HIcor_Bsum = 0
    InitCond.Fpol = 0
    InitCond.sCor1 = 0
    InitCond.sCor2 = 0

    # Growth stage
    InitCond.GrowthStage = 0

    # Transpiration
    InitCond.TrRatio = 1

    # crop growth
    InitCond.rCor = 1

    InitCond.CC = 0
    InitCond.CCadj = 0
    InitCond.CC_NS = 0
    InitCond.CCadj_NS = 0
    InitCond.B = 0
    InitCond.B_NS = 0
    InitCond.HI = 0
    InitCond.HIadj = 0
    InitCond.CCxAct = 0
    InitCond.CCxAct_NS = 0
    InitCond.CCxW = 0
    InitCond.CCxW_NS = 0
    InitCond.CCxEarlySen = 0
    InitCond.CCprev = 0
    InitCond.ProtectedSeed = 0

    ## Update CO2 concentration ##
    # Get CO2 concentration
    Yri = pd.DatetimeIndex([ClockStruct.StepStartTime]).year[0]
    CO2.CurrentConc = CO2_data.loc[Yri]
    # Get CO2 weighting factor for first year
    CO2conc = CO2.CurrentConc
    CO2ref = CO2.RefConc
    if CO2conc <= CO2ref:
        fw = 0
    else:
        if CO2conc >= 550:
            fw = 1
        else:
            fw = 1-((550-CO2conc)/(550-CO2ref))


    # Determine initial adjustment
    fCO2 = (CO2conc/CO2ref)/(1+(CO2conc-CO2ref)\
                             *((1-fw)*Crop.bsted+fw*((Crop.bsted*Crop.fsink)\
                                                     +(Crop.bface*(1-Crop.fsink)))))

    # Consider crop type
    if Crop.WP >= 40:
        # No correction for C4 crops
        ftype = 0
    elif Crop.WP <= 20:
        # Full correction for C3 crops
        ftype = 1
    else:
        ftype = (40-Crop.WP)/(40-20)

    # Total adjustment
    Crop.fCO2 = 1+ftype*(fCO2-1)


    ## Reset soil water conditions (if not running off-season) ##
    if ClockStruct.SimOffSeason=='N':
        # Reset water content to starting conditions
        InitCond.th = InitCond.thini
        # Reset surface storage
        if (FieldMngt.Bunds == 'Y') and (FieldMngt.zBund > 0.001):
            # Get initial storage between surface bunds
            InitCond.SurfaceStorage = min(FieldMngt.BundWater,FieldMngt.zBund)
        else:
            # No surface bunds
            InitCond.SurfaceStorage = 0


    weather_df = InitCond.weather
    ## Update crop parameters (if in GDD mode) ##
    if Crop.CalendarType == 2:
        # Extract weather data for upcoming growing season
        wdf = weather_df[weather_df[:,4]>=ClockStruct.PlantingDates[ClockStruct.SeasonCounter]]
        wdf = wdf[wdf[:,4]<=ClockStruct.HarvestDates[ClockStruct.SeasonCounter]]
        Tmin = wdf[:,0]
        Tmax = wdf[:,1]

        # Calculate GDD's
        if Crop.GDDmethod == 1:
            Tmean = (Tmax+Tmin)/2
            Tmean[Tmean>Crop.Tupp] = Crop.Tupp
            Tmean[Tmean<Crop.Tbase] = Crop.Tbase
            GDD = Tmean-Crop.Tbase
        elif Crop.GDDmethod == 2:
            Tmax[Tmax>Crop.Tupp] = Crop.Tupp
            Tmax[Tmax<Crop.Tbase] = Crop.Tbase
            Tmin[Tmin>Crop.Tupp] = Crop.Tupp
            Tmin[Tmin<Crop.Tbase] = Crop.Tbase
            Tmean = (Tmax+Tmin)/2
            GDD = Tmean-Crop.Tbase
        elif Crop.GDDmethod == 3:
            Tmax[Tmax>Crop.Tupp] = Crop.Tupp
            Tmax[Tmax<Crop.Tbase] = Crop.Tbase
            Tmin[Tmin>Crop.Tupp] = Crop.Tupp
            Tmean = (Tmax+Tmin)/2
            Tmean[Tmean<Crop.Tbase] = Crop.Tbase
            GDD = Tmean-Crop.Tbase



        GDDcum = np.cumsum(GDD)




        # 1. GDD's from sowing to maximum canopy cover
        Crop.MaxCanopyCD = (GDDcum>Crop.MaxCanopy).argmax()+1
        # 2. GDD's from sowing to end of vegetative growth
        Crop.CanopyDevEndCD = (GDDcum>Crop.CanopyDevEnd).argmax()+1
        # 3. Calendar days from sowing to start of yield formation
        Crop.HIstartCD = (GDDcum>Crop.HIstart).argmax()+1
        # 4. Calendar days from sowing to end of yield formation
        Crop.HIendCD = (GDDcum>Crop.HIend).argmax()+1
        # 5. Duration of yield formation in calendar days
        Crop.YldFormCD = Crop.HIendCD-Crop.HIstartCD
        if Crop.CropType == 3:
            # 1. Calendar days from sowing to end of flowering
            FloweringEnd = (GDDcum>Crop.FloweringEnd).argmax()+1
            # 2. Duration of flowering in calendar days
            Crop.FloweringCD = FloweringEnd-Crop.HIstartCD




        # Update harvest index growth coefficient
        Crop = calculate_HIGC(Crop)

        # Update day to switch to linear HI build-up
        if Crop.CropType == 3:
            # Determine linear switch point and HIGC rate for fruit/grain crops
            Crop = calculate_HI_linear(Crop)

        else:
            # No linear switch for leafy vegetable or root/tiber crops
            Crop.tLinSwitch = []
            Crop.dHILinear = []



    ## Update global variables ##
    InitCond.ParamStruct.Seasonal_Crop_List[ClockStruct.SeasonCounter] = Crop
    InitCond.ParamStruct.CO2 = CO2

    return InitCond

# Cell
def update_time(ClockStruct,InitCond,Outputs):
    """
    Function to update current time in model

    *Arguments:*\n

    `ClockStruct` : `ClockStructClass` :  model time paramaters

    `InitCond` : `InitCondClass` :  containing current model paramaters

    `weather_df`: `pandas.DataFrame` :  weather data for simulation period


    *Returns:*

    `ClockStruct` : `ClockStructClass` :  model time paramaters


    `InitCond` : `InitCondClass` :  containing reset model paramaters


    """
    ## Update time ##
    if ClockStruct.ModelTermination == False:
        if  (InitCond.HarvestFlag == True) \
        and ((ClockStruct.SimOffSeason=='N')):
            # End of growing season has been reached and not simulating
            # off-season soil water balance. Advance time to the start of the
            # next growing season.
            # Check if in last growing season
            if ClockStruct.SeasonCounter < ClockStruct.nSeasons-1:
                # Update growing season counter
                ClockStruct.SeasonCounter = ClockStruct.SeasonCounter+1
                # Update time-step counter
                #ClockStruct.TimeSpan = pd.Series(ClockStruct.TimeSpan)
                ClockStruct.TimeStepCounter = ClockStruct.TimeSpan.get_loc(ClockStruct.PlantingDates[ClockStruct.SeasonCounter])
                # Update start time of time-step
                ClockStruct.StepStartTime = ClockStruct.TimeSpan[ClockStruct.TimeStepCounter]
                # Update end time of time-step
                ClockStruct.StepEndTime = ClockStruct.TimeSpan[ClockStruct.TimeStepCounter + 1]
                # Reset initial conditions for start of growing season
                InitCond =  reset_initial_conditions(ClockStruct,InitCond)

        else:
            # Simulation considers off-season, so progress by one time-step
            # (one day)
            # Time-step counter
            ClockStruct.TimeStepCounter = ClockStruct.TimeStepCounter+1
            # Start of time step (beginning of current day)
            #ClockStruct.TimeSpan = pd.Series(ClockStruct.TimeSpan)
            ClockStruct.StepStartTime = ClockStruct.TimeSpan[ClockStruct.TimeStepCounter]
            # End of time step (beginning of next day)
            ClockStruct.StepEndTime = ClockStruct.TimeSpan[ClockStruct.TimeStepCounter + 1]
            # Check if in last growing season
            if ClockStruct.SeasonCounter < ClockStruct.nSeasons-1:
                # Check if upcoming day is the start of a new growing season
                if ClockStruct.StepStartTime == ClockStruct.PlantingDates[ClockStruct.SeasonCounter+1]:
                    # Update growing season counter
                    ClockStruct.SeasonCounter = ClockStruct.SeasonCounter+1
                    # Reset initial conditions for start of growing season
                    InitCond =  reset_initial_conditions(ClockStruct,InitCond)


    elif ClockStruct.ModelTermination == True:
        ClockStruct.StepStartTime = ClockStruct.StepEndTime
        ClockStruct.StepEndTime = ClockStruct.StepEndTime + np.timedelta64(1, 'D')

        Outputs.Flux = pd.DataFrame(Outputs.Flux, columns=["TimeStepCounter",\
                                          "SeasonCounter","DAP","Wr","zGW",\
                                          "SurfaceStorage","IrrDay",\
                                          "Infl","Runoff","DeepPerc","CR",\
                                          "GwIn","Es","EsPot","Tr","P"])


        Outputs.Water =pd.DataFrame(Outputs.Water, columns=["TimeStepCounter","GrowingSeason","DAP"]\
                                    +['th'+str(i) for i in range(1,Outputs.Water.shape[1]-2)])

        Outputs.Growth = pd.DataFrame(Outputs.Growth, columns = ["TimeStepCounter",'SeasonCounter',"DAP",'GDD',\
                                         'GDDcum','Zroot',\
                                         'CC','CC_NS','B',\
                                         'B_NS','HI','HIadj',\
                                         'Y'])


    return ClockStruct,InitCond,Outputs






# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/00_core.ipynb (unless otherwise specified).

__all__ = ['AquaCropModel', 'perform_timestep']

# Cell
import matplotlib.pyplot as plt
import numpy as np
import pathlib
import os
import pandas as pd
import sys
[sys.path.append(i) for i in ['.', '..']]

# Cell
from .initialize import *
from .timestep import *
from .classes import *

# Cell
class AquaCropModel:

    def __init__(self,SimStartTime,SimEndTime,wdf,Soil,Crop,
                     Farmer=None,FieldMngt=None,FallowFieldMngt=None,
                     GwStruct=None,InitWC=None,planting_dates=None,
                     harvest_dates=None):

        self.SimStartTime = SimStartTime
        self.SimEndTime = SimEndTime
        self.wdf = wdf
        self.Soil = Soil
        self.Crop = Crop
        self.planting_dates = planting_dates
        self.harvest_dates = harvest_dates

        self.Farmer = Farmer
        self.FieldMngt = FieldMngt
        self.FallowFieldMngt = FallowFieldMngt
        self.GwStruct = GwStruct
        self.InitWC = InitWC

        if Farmer == None:  self.Farmer = FarmerClass(IrrMethod=0);
        if FieldMngt == None:  self.FieldMngt = FieldMngtClass();
        if FallowFieldMngt == None:  self.FallowFieldMngt = FieldMngtClass();
        if GwStruct == None:  self.GwStruct = GwClass();
        if InitWC == None:  self.InitWC = InitWCClass();



    def initialize(self,):
        """
        Initialize variables

        Looks for filepaths within aquacrop/data/`fileDir`/Input/FileSetup.txt
        and stores inside a `FileLocationsClass`

        """

        # define model runtime
        self.ClockStruct = read_clock_paramaters(self.SimStartTime,self.SimEndTime)

        # get weather data
        self.weather_df = read_weather_inputs(self.ClockStruct,self.wdf)

        # read model params
        self.ClockStruct, self.ParamStruct = read_model_parameters(self.ClockStruct,self.Soil,
                                                                   self.Crop,self.planting_dates,self.harvest_dates)

        # read irrigation management
        self.ParamStruct,self.Farmer = read_irrigation_management(self.ParamStruct,self.Farmer,self.ClockStruct)

        # read field management
        self.ParamStruct = read_field_management(self.ParamStruct,self.FieldMngt,self.FallowFieldMngt)

        # read groundwater table
        self.ParamStruct = read_groundwater_table(self.ParamStruct,self.GwStruct,self.ClockStruct)

        # Compute additional variables
        self.ParamStruct = compute_variables(self.ParamStruct,self.weather_df,self.ClockStruct)

        # read, calculate inital conditions
        self.ParamStruct, self.InitCond = read_model_initial_conditions(self.ParamStruct,self.ClockStruct,self.InitWC)

        self.ParamStruct = create_soil_profile(self.ParamStruct)

        self.InitCond.ParamStruct = self.ParamStruct

        Outputs = OutputClass()
        Outputs.Water = np.zeros((len(self.ClockStruct.TimeSpan),15))
        Outputs.Flux = np.zeros((len(self.ClockStruct.TimeSpan),16))
        Outputs.Growth = np.zeros((len(self.ClockStruct.TimeSpan),13))
        Outputs.Final = pd.DataFrame(columns = ['Season','Name','HarvestDate','Step','Yield','Tirr'])


        self.Outputs=Outputs

        # save model weather to InitCond
        self.InitCond.weather = self.weather_df.values



        #return self.ClockStruct,self.InitCond,self.Outputs
        return


    def step(self,num_steps=1,till_termination=False):

        if till_termination==True:

            while self.ClockStruct.ModelTermination == False:

                self.ClockStruct,self.InitCond,self.Outputs = perform_timestep(self.ClockStruct,
                                                                               self.InitCond,
                                                                               self.Outputs,
                                                                               self.Farmer)
        else:

            for i in range(num_steps):

                self.ClockStruct,self.InitCond,self.Outputs = perform_timestep(self.ClockStruct,
                                                                               self.InitCond,
                                                                               self.Outputs,
                                                                               self.Farmer)

                if self.ClockStruct.ModelTermination: return

        #return self.ClockStruct,self.InitCond,self.Outputs
        return

# Cell
def perform_timestep(ClockStruct,InitCond,Outputs,Farmer):

    """
    Function to run a single time-step (day) calculation of AquaCrop-OS

    *Arguments:*\n


    `ClockStruct` : `ClockStructClass` :  model time paramaters

    `weather_df`: `pandas.DataFrame` :  weather data for simulation period

    `InitCond` : `InitCondClass` :  containing all initiall model paramaters

    `Outputs` : `OutputClass` :  object to store outputs


    *Returns:*

    `ClockStruct` : `ClockStructClass` :  updated time paramaters

    `InitCond` : `InitCondClass` :  updated model paramaters

    `Outputs` : `OutputClass` :  updated with the days outputs



    """


    # extract weather data for current timestep
    #weather_step = weather_df[weather_df.Date==ClockStruct.StepStartTime]
    weather_step = InitCond.weather[ClockStruct.TimeStepCounter]

    #%% Get model solution %%
    NewCond,Outputs = solution(InitCond,ClockStruct,weather_step,Outputs,Farmer)

    #%% Check model termination %%
    ClockStruct = check_model_termination(ClockStruct,NewCond)

    #%% Update time step %%
    ClockStruct, InitCond,Outputs = update_time(ClockStruct,NewCond,Outputs)

    return ClockStruct,InitCond,Outputs


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/05_comparison.ipynb (unless otherwise specified).

__all__ = ['run_comparison']

# Cell
import sys
_=[sys.path.append(i) for i in ['.', '..']]

# Cell
def run_comparison(model,name):
    """
    Function to run a comparison between python matlab and windows.
    Plots yields and prints mean and mean absolute error between them

    *Arguments:*


    `name`: `str` : name of directory containing input files

    *Returns:*

    None


    """
    Outputs = model.Outputs

    py = Outputs.Final.round(3)

    matlab = pd.read_csv('../aquacrop/data/'+name+'_matlab.txt',delim_whitespace=True,header=None)
    matlab.columns = ["season","crop","plantdate","stepplant","harvestdate","stepharvest","Yield","tirr"]


    windows_names = ('    RunNr     Day1   Month1    Year1     Rain      ETo       GD     CO2      Irri   Infilt   Runoff    Drain   Upflow        E     E/Ex       Tr      TrW   Tr/Trx    SaltIn   SaltOut    SaltUp  SaltProf     Cycle   SaltStr  FertStr  WeedStr  TempStr   ExpStr   StoStr  BioMass  Brelative   HI     Yield     WPet     DayN   MonthN    YearN'.split())
    windows = pd.read_csv('../aquacrop/data/'+name+'_windows.OUT',skiprows=5,delim_whitespace=True,names=windows_names)

    combined = pd.DataFrame([py.Yield,windows.Yield,matlab.Yield]).T

    combined.columns = ["py", "windows","matlab"]
    mae = np.round(np.abs(combined.py - combined.windows).mean(),2)
    pymean = combined.mean().py.round(2)
    print(f'python seasonal mean: {pymean} kg/ha\nMAE from windows: {mae} kg/ha')

    mae_mat = np.round(np.abs(combined.py - combined.matlab).mean(),3)
    print(f'MAE from matlab:  {mae_mat} kg/ha')

    plt.style.use('seaborn')

    fig, ax = plt.subplots(2,1,sharex=True,figsize=(11,8))

    ax[0].plot(py.Yield,label='Python')
    ax[0].plot(matlab.Yield,label='Matlab')
    ax[0].plot(windows.Yield,'--',label='Windows')
    ax[0].legend(fontsize = 18)
    ax[0].set_ylabel('Yield',fontsize=18)

    #sns.jointplot(np.arange(len(py)), py.Yield - windows.Yield,
      #            kind="resid",color="m",ratio=10)

    ax[1].scatter(np.arange(len(py)),py.Yield - windows.Yield,label='Python')
    ax[1].scatter(np.arange(len(py)),matlab.Yield - windows.Yield,label='Matlab')
    ax[1].plot([0,len(py)],[0,0],'--',color='black')
    ax[1].set_xlabel('Season',fontsize=18)
    ax[1].set_ylabel('Residuals',fontsize=18)
    ax[1].legend(fontsize = 18)





    plt.show()


    return Outputs,windows


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/01_initialize.ipynb (unless otherwise specified).

__all__ = ['prepare_weather', 'read_clock_paramaters', 'read_weather_inputs', 'read_model_parameters',
           'read_irrigation_management', 'read_field_management', 'read_groundwater_table', 'compute_variables',
           'compute_crop_calander', 'calculate_HIGC', 'calculate_HI_linear', 'read_model_initial_conditions',
           'create_soil_profile']

# Cell
import numpy as np
import os
import pandas as pd
from .classes import *
import pathlib
import types
from copy import deepcopy
from collections.abc import Iterable
import aquacrop


# Cell
def prepare_weather(weatherFilePath):
    """
    function to read in weather data and return a dataframe containing
    the weather data

    *Arguments:*\n

    `FileLocations` : `FileLocationsClass`:  input File Locations

    `ClockStruct` : `ClockStructClass` :  model time paramaters

    `weatherFilePath` : `str` :  file location of weather data



    *Returns:*

    `weather_df`: `pandas.DataFrame` :  weather data for simulation period

    """


    weather_df = pd.read_csv(weatherFilePath,header=0,delim_whitespace=True)

    assert len(weather_df.columns) == 7

    # rename the columns
    weather_df.columns = str("Day Month Year MinTemp MaxTemp Precipitation ReferenceET").split()

    # put the weather dates into datetime format
    weather_df["Date"] = pd.to_datetime(weather_df[['Year','Month','Day']])

    # drop the day month year columns
    weather_df = weather_df.drop(["Day","Month","Year"],axis=1)


    return weather_df

# Cell
def read_clock_paramaters(SimStartTime,SimEndTime,OffSeason='N'):
    """
    function to read in clock paramaters and return a `ClockStructClass` object



    """


    # extract data and put into numpy datetime format
    SimStartTime = pd.to_datetime(SimStartTime)
    SimEndTime = pd.to_datetime(SimEndTime)


    # create object
    ClockStruct = ClockStructClass()

    # add variables
    ClockStruct.SimulationStartDate = SimStartTime
    ClockStruct.SimulationEndDate = SimEndTime

    ClockStruct.nSteps =  (SimEndTime - SimStartTime).days + 1
    ClockStruct.TimeSpan = pd.date_range(freq='D',start=SimStartTime,end=SimEndTime)

    ClockStruct.StepStartTime = ClockStruct.TimeSpan[0]
    ClockStruct.StepEndTime = ClockStruct.TimeSpan[1]

    ClockStruct.SimOffSeason = OffSeason

    return ClockStruct

# Cell
def read_weather_inputs(ClockStruct,weather_df):
    """
    function to read in weather data and return a dataframe containing
    the weather data file for the simulation period only


    """



        # get the start and end dates of simulation
    start_date = ClockStruct.SimulationStartDate
    end_date = ClockStruct.SimulationEndDate

    assert weather_df.Date.iloc[0] <= start_date
    assert weather_df.Date.iloc[-1] >= end_date

    # remove weather data outside of simulation dates
    weather_df = weather_df[weather_df.Date>=start_date]
    weather_df = weather_df[weather_df.Date<=end_date]




    return weather_df

# Cell
def read_model_parameters(ClockStruct,Soil,Crop,planting_dates,harvest_dates):
    """
    Function to read input files and initialise soil and crop parameters


    *Arguments:*\n

    `FileLocations` : `FileLocationsClass`:  input File Locations

    `ClockStruct` : `ClockStructClass` :  model time paramaters

    *Returns:*

    `FileLocations` : `FileLocationsClass`: updated input File Locations

    `ClockStruct` : `ClockStructClass` : updated time paramaters

    `ParamStruct` : `ParamStructClass` :  Contains model crop and soil paramaters

    """

    # create ParamStruct object
    ParamStruct = ParamStructClass()


    Soil.fill_nan()


    # Assign Soil object to ParamStruct
    ParamStruct.Soil = Soil





    ###########
    # crop
    ###########



    if isinstance(Crop, Iterable):
        CropList=list(Crop)
    else:
        CropList = [Crop]


    # assign variables to paramstruct
    ParamStruct.NCrops = len(CropList)
    if ParamStruct.NCrops > 1:
        ParamStruct.SpecifiedPlantCalander = 'Y'
    else:
        ParamStruct.SpecifiedPlantCalander = 'N'



    #Crop.Ksw = pd.Series({'Exp':1.,'Sto':1,'Sen':1,'Pol':1,'StoLin':1})


    for crop in CropList:
        assert Soil.zSoil > crop.Zmax, 'max rooting depth exceeds soil profile'


    # add crop list to ParamStruct
    ParamStruct.CropList = CropList

    ############################
    # plant and harvest times
    ############################


    # find planting and harvest dates
    # check if there is more than 1 crop or multiple plant dates in sim year
    if ParamStruct.SpecifiedPlantCalander == "Y":
        # if here than crop rotation occours during same period

        # create variables from dataframe
        PlantingDates = pd.to_datetime(planting_dates)
        HarvestDates = pd.to_datetime(harvest_dates)


        if (ParamStruct.NCrops > 1):

            CropChoices = [crop.Name for crop in ParamStruct.CropList]


        assert len(CropChoices) == len(PlantingDates) == len(HarvestDates)




    elif ParamStruct.NCrops == 1:
        # Only one crop type considered during simulation - i.e. no rotations
        # either within or between years

        # Get start and end years for full simulation
        SimStartDate = ClockStruct.SimulationStartDate
        SimEndDate = ClockStruct.SimulationEndDate

        # extract the years and months of these dates
        start_end_years = pd.DatetimeIndex([SimStartDate,SimEndDate]).year
        start_end_months = pd.DatetimeIndex([SimStartDate,SimEndDate]).month

        # check if crop growing season runs over calander year
        # Planting and harvest dates are in days/months format so just add arbitrary year
        singleYear = pd.to_datetime("1990/"+Crop.PlantingDate) < pd.to_datetime("1990/"+Crop.HarvestDate)
        if singleYear:
            # if normal year

            # specify the planting and harvest years as normal
            plant_years = list(range(start_end_years[0],start_end_years[1]+1))
            harvest_years = plant_years
        else:
            # if it takes over a year then the plant year finishes 1 year before end of sim
            # and harvest year starts 1 year after sim start

            if pd.to_datetime(str(start_end_years[1]+2)+'/'+Crop.HarvestDate) < SimEndDate:

                # specify shifted planting and harvest years
                plant_years = list(range(start_end_years[0],start_end_years[1]+1))
                harvest_years = list(range(start_end_years[0]+1,start_end_years[1]+2))
            else:

                plant_years = list(range(start_end_years[0],start_end_years[1]))
                harvest_years = list(range(start_end_years[0]+1,start_end_years[1]+1))



        # Correct for partial first growing season (may occur when simulating
        # off-season soil water balance)
        if pd.to_datetime(str(plant_years[0])+"/"+Crop.PlantingDate) < ClockStruct.SimulationStartDate:
            # shift everything by 1 year
            plant_years = plant_years[1:]
            harvest_years = harvest_years[1:]



        # ensure number of planting and harvest years are the same
        assert len(plant_years) == len(harvest_years)

        # create lists to hold variables
        PlantingDates = []
        HarvestDates = []
        CropChoices = []

        # save full harvest/planting dates and crop choices to lists
        for i in range(len(plant_years)):
            PlantingDates.append(str(plant_years[i])  + "/" + ParamStruct.CropList[0].PlantingDate )
            HarvestDates.append(str(harvest_years[i])  + "/" + ParamStruct.CropList[0].HarvestDate )
            CropChoices.append( ParamStruct.CropList[0].Name )


    # save crop choices
    ParamStruct.CropChoices = list(CropChoices)


    # save clock paramaters
    ClockStruct.PlantingDates = pd.to_datetime(PlantingDates)
    ClockStruct.HarvestDates = pd.to_datetime(HarvestDates)
    ClockStruct.nSeasons = len(PlantingDates)

    # Initialise growing season counter
    if pd.to_datetime(ClockStruct.StepStartTime) == ClockStruct.PlantingDates[0]:
        ClockStruct.SeasonCounter = 0
    else:
        ClockStruct.SeasonCounter = -1

    # return the FileLocations object as i have added some elements
    return ClockStruct, ParamStruct

# Cell
def read_irrigation_management(ParamStruct,Farmer,ClockStruct):
    """
    Function to read and initialise irrigation management parameters




    """

    for i in range(ParamStruct.NCrops):

        # get crop object
        Crop = ParamStruct.CropList[i]



        Crop.IrrMngt = Farmer


        # If specified, read input irrigation time-series
        if Crop.IrrMngt.IrrMethod == 3:

            df = Farmer.Schedule

            # change the index to the date
            df.index = pd.DatetimeIndex(df.Date)

            # create a dateframe containing the daily irrigation to
            # be applied for every day in the simulation
            df = df.reindex(ClockStruct.TimeSpan, fill_value=0).drop("Date",axis=1)

            Farmer.Schedule = df

            # assign to Crop object
            Crop.IrrSchd = df



        ParamStruct.CropList[i] = Crop


    return ParamStruct,Farmer

# Cell
def read_field_management(ParamStruct,FieldMngt,FallowFieldMngt):
    """
    Function to read input files and initialise field management parameters

    *Arguments:*\n

    `ParamStruct` : `ParamStructClass` :  Contains model crop and soil paramaters

    `FileLocations` : `FileLocationsClass`:  input File Locations

    *Returns:*

    `ParamStruct` : `ParamStructClass` :  updated with field management info


    """


    for i in range(ParamStruct.NCrops):

        # get crop object
        Crop = ParamStruct.CropList[i]

        Crop.FieldMngt = FieldMngt

        ParamStruct.CropList[i] = Crop


    ParamStruct.FallowFieldMngt = FallowFieldMngt

    return ParamStruct

# Cell
def read_groundwater_table(ParamStruct,GwStruct,ClockStruct):
    """
    Function to read input files and initialise groundwater parameters


    """


    # assign water table value and method
    WT = GwStruct.WaterTable
    WTMethod = GwStruct.Method


    # check if water table present
    if WT == 'N':
        ParamStruct.WaterTable = 0

    elif WT == 'Y':
        ParamStruct.WaterTable = 1


        df=pd.DataFrame([GwStruct.dates,GwStruct.values]).T
        df.columns=['Date','Depth(mm)']


        # get date in correct format
        df.Date = pd.DatetimeIndex(df.Date)


        if len(df)==1:

            # if only 1 watertable depth then set that value to be constant
            # accross whole simulation
            zGW = df.reindex(ClockStruct.TimeSpan,
                             fill_value=df["Depth(mm)"].iloc[0],
                            ).drop("Date",axis=1)["Depth(mm)"]

        elif len(df)>1:
            # check water table method
            if WTMethod == "Constant":
                # No interpolation between dates

                # create daily depths for each simulation day
                zGW = pd.Series(np.nan*np.ones(len(ClockStruct.TimeSpan)),\
                                index = ClockStruct.TimeSpan)


                # assign constant depth for all dates in between
                for row in range(len(df)):
                    date = df.Date.iloc[row]; depth = df["Depth(mm)"].iloc[row]
                    zGW.loc[zGW.index>=date] = depth
                    if row == 0:
                        zGW.loc[zGW.index<=date] = depth



            elif WTMethod == "Variable":
                # Linear interpolation between dates

                # create daily depths for each simulation day
                # fill unspecified days with NaN
                zGW = pd.Series(np.nan*np.ones(len(ClockStruct.TimeSpan)),\
                                index = ClockStruct.TimeSpan)


                for row in range(len(df)):
                    date = df.Date.iloc[row]; depth = df["Depth(mm)"].iloc[row]
                    zGW.loc[date] = depth

                # Interpolate daily groundwater depths
                zGW = zGW.interpolate()


        # assign values to Paramstruct object
        ParamStruct.zGW = zGW
        ParamStruct.WTMethod = WTMethod


    return ParamStruct

# Cell
def compute_variables(ParamStruct,weather_df,ClockStruct,acfp=pathlib.Path(os.path.abspath(aquacrop.__file__)).parent):
    """
    Function to compute additional variables needed to run the model



    """


    if ParamStruct.WaterTable == 1:

        ParamStruct.Soil.add_capillary_rise_params()



    # Calculate readily evaporable water in surface layer
    if ParamStruct.Soil.AdjREW == 0:
        ParamStruct.Soil.REW = round((1000*(ParamStruct.Soil.profile.th_fc.iloc[0]
                                            - ParamStruct.Soil.profile.th_dry.iloc[0])
                                            * ParamStruct.Soil.EvapZsurf),2)

    if ParamStruct.Soil.CalcCN == 1:
        # adjust curve number
        ksat = ParamStruct.Soil.profile.Ksat.iloc[0]
        if ksat > 864:
            ParamStruct.Soil.CN = 46
        elif ksat > 347:
            ParamStruct.Soil.CN = 61
        elif ksat > 36:
            ParamStruct.Soil.CN = 72
        elif ksat > 0:
            ParamStruct.Soil.CN = 77

        assert ksat >0


    for i in range(ParamStruct.NCrops):

        crop = ParamStruct.CropList[i]
        crop.calculate_additional_params()




        # Crop calander
        crop = compute_crop_calander(crop,ParamStruct,ClockStruct,weather_df)

        # Harvest index ParamStruct.Seasonal_Crop_List[ClockStruct.SeasonCounter].Paramsgrowth coefficient
        crop = calculate_HIGC(crop)

        # Days to linear HI switch point
        if crop.CropType == 3:
            # Determine linear switch point and HIGC rate for fruit/grain crops
            crop = calculate_HI_linear(crop)
        else:
            # No linear switch for leafy vegetable or root/tiber crops
            crop.tLinSwitch = []
            crop.dHILinear = []


        ParamStruct.CropList[i] = crop



    ## Calculate WP adjustment factor for elevation in CO2 concentration ##
    # Load CO2 data
    co2Data = pd.read_csv(acfp/'data/MaunaLoaCO2.txt',header=1,delim_whitespace=True,names=["year","ppm"])

    # Years
    start_year, end_year = pd.DatetimeIndex([ClockStruct.SimulationStartDate,ClockStruct.SimulationEndDate]).year
    sim_years = np.arange(start_year,end_year+1)

    ParamStruct.CO2 = CO2Class()

    # Interpolate data
    CO2conc = np.interp(sim_years,co2Data.year,co2Data.ppm)

    # Store data
    ParamStruct.CO2data = pd.Series(CO2conc,index=sim_years) # maybe get rid of this

    # Define reference CO2 concentration
    ParamStruct.CO2.RefConc = 369.41


    # Get CO2 concentration for first year
    ParamStruct.CO2.CurrentConc = ParamStruct.CO2data.iloc[0]

    # Get CO2 weighting factor for first year
    CO2ref = ParamStruct.CO2.RefConc
    CO2conc = ParamStruct.CO2.CurrentConc
    if CO2conc <= CO2ref:
        fw = 0
    else:
        if CO2conc >= 550:
            fw = 1
        else:
            fw = 1-((550-CO2conc)/(550-CO2ref))


    # Determine adjustment for each crop in first year of simulation
    for i in range(ParamStruct.NCrops):
        crop = ParamStruct.CropList[i]
        # Determine initial adjustment
        fCO2 = (CO2conc/CO2ref)/(1+(CO2conc-CO2ref)*((1-fw)*crop.bsted+fw*((crop.bsted*crop.fsink)
                                                                           +(crop.bface*(1-crop.fsink)))))

        # Consider crop type
        if crop.WP >= 40:
            #No correction for C4 crops
            ftype = 0
        elif crop.WP <= 20:
        # Full correction for C3 crops
            ftype = 1
        else:
            ftype = (40-crop.WP)/(40-20)

        # Total adjustment
        crop.fCO2 = 1+ftype*(fCO2-1)

        ParamStruct.CropList[i] = crop



    # change this later
    if ParamStruct.NCrops == 1:
        ParamStruct.Seasonal_Crop_List = [deepcopy(ParamStruct.CropList[0]) for i in range(len(ParamStruct.CropChoices))]

    else:
        ParamStruct.Seasonal_Crop_List = ParamStruct.CropList

    # add crop for out of growing season
    ParamStruct.Fallow_Crop = deepcopy(ParamStruct.Seasonal_Crop_List[0])




    return ParamStruct

# Cell
def compute_crop_calander(crop,ParamStruct,ClockStruct,weather_df):
    """
    Function to compute additional parameters needed to define crop phenological calendar



    *Arguments:*\n

    `crop` : `CropClass` :  Crop object containing crop paramaters


    `ParamStruct` : `ParamStructClass` :  Contains model crop and soil paramaters

    `ClockStruct` : `ClockStructClass` :  model time paramaters

    `weather_df`: `pandas.DataFrame` :  weather data for simulation period


    *Returns:*

    `crop` : `CropClass` : updated Crop object



    """



    #Define crop calendar mode
    Mode = crop.CalendarType

    #Calculate variables %%
    if Mode == 1: # Growth in calendar days
        # Time from sowing to end of vegatative growth period
        if crop.Determinant == 1:
            crop.CanopyDevEnd = round(crop.HIstart+(crop.Flowering/2))
        else:
            crop.CanopyDevEnd = crop.Senescence

        # Time from sowing to 10% canopy cover (non-stressed conditions)
        crop.Canopy10Pct = round(crop.Emergence+(np.log(0.1/crop.CC0)/crop.CGC))


        # Time from sowing to maximum canopy cover (non-stressed conditions)
        crop.MaxCanopy = round( crop.Emergence
                                          +(  np.log( (0.25*crop.CCx*crop.CCx/crop.CC0)
                                                      /(crop.CCx-(0.98*crop.CCx))
                                                    )
                                              /crop.CGC
                                           )
                                        )

        # Time from sowing to end of yield formation
        crop.HIend = crop.HIstart+crop.YldForm


        # Duplicate calendar values (needed to minimise if statements when switching between GDD and CD runs)
        crop.EmergenceCD = crop.Emergence
        crop.Canopy10PctCD = crop.Canopy10Pct
        crop.MaxRootingCD = crop.MaxRooting
        crop.SenescenceCD = crop.Senescence
        crop.MaturityCD = crop.Maturity
        crop.MaxCanopyCD = crop.MaxCanopy
        crop.CanopyDevEndCD = crop.CanopyDevEnd
        crop.HIstartCD = crop.HIstart
        crop.HIendCD = crop.HIend
        crop.YldFormCD = crop.YldForm
        if crop.CropType == 3:
            crop.FloweringEnd = crop.HIstart+crop.Flowering
            crop.FloweringEndCD = crop.FloweringEnd
            crop.FloweringCD = crop.Flowering


        # Check if converting crop calendar to GDD mode
        if crop.SwitchGDD == 1:
            # Extract weather data for first growing season that crop is planted
            for i,n in enumerate(ParamStruct.CropChoices):
                if n == crop.Name:
                    idx = i
                    break
                else:
                    idx = -1
            assert idx > -1

            date_range = pd.date_range(ClockStruct.PlantingDates[idx],ClockStruct.HarvestDates[idx])
            wdf = weather_df.copy(); wdf.index = wdf.Date
            wdf = wdf.loc[date_range]
            Tmin = wdf.MinTemp
            Tmax = wdf.MaxTemp


            # Calculate GDD's
            if crop.GDDmethod == 1:

                Tmean = (Tmax+Tmin)/2
                Tmean = Tmean.clip(lower=crop.Tbase,upper=crop.Tupp)
                GDD = Tmean-crop.Tbase

            elif crop.GDDmethod == 2:

                Tmax = Tmax.clip(lower=crop.Tbase,upper=crop.Tupp)
                Tmin = Tmin.clip(lower=crop.Tbase,upper=crop.Tupp)
                Tmean = (Tmax+Tmin)/2
                GDD = Tmean-crop.Tbase

            elif crop.GDDmethod == 3:

                Tmax = Tmax.clip(lower=crop.Tbase,upper=crop.Tupp)
                Tmin = Tmin.clip(upper=crop.Tupp)
                Tmean = (Tmax+Tmin)/2
                Tmean = Tmean.clip(lower=crop.Tbase)
                GDD = Tmean-crop.Tbase

            GDDcum = np.cumsum(GDD)
            # Find GDD equivalent for each crop calendar variable
            # 1. GDD's from sowing to emergence
            crop.Emergence = GDDcum.iloc[int(crop.EmergenceCD)]
            # 2. GDD's from sowing to 10# canopy cover
            crop.Canopy10Pct = GDDcum.iloc[int(crop.Canopy10PctCD)]
            # 3. GDD's from sowing to maximum rooting
            crop.MaxRooting = GDDcum.iloc[int(crop.MaxRootingCD)]
            # 4. GDD's from sowing to maximum canopy cover
            crop.MaxCanopy = GDDcum.iloc[int(crop.MaxCanopyCD)]
            # 5. GDD's from sowing to end of vegetative growth
            crop.CanopyDevEnd = GDDcum.iloc[int(crop.CanopyDevEndCD)]
            # 6. GDD's from sowing to senescence
            crop.Senescence = GDDcum.iloc[int(crop.SenescenceCD)]
            # 7. GDD's from sowing to maturity
            crop.Maturity = GDDcum.iloc[int(crop.MaturityCD)]
            # 8. GDD's from sowing to start of yield formation
            crop.HIstart = GDDcum.iloc[int(crop.HIstartCD)]
            # 9. GDD's from sowing to start of yield formation
            crop.HIend = GDDcum.iloc[int(crop.HIendCD)]
            # 10. Duration of yield formation (GDD's)
            crop.YldForm = crop.HIend-crop.HIstart

            # 11. Duration of flowering (GDD's) - (fruit/grain crops only)
            if crop.CropType == 3:
                # GDD's from sowing to end of flowering
                crop.FloweringEnd = GDDcum.iloc[int(crop.FloweringEndCD)]
                # Duration of flowering (GDD's)
                crop.Flowering = crop.FloweringEnd-crop.HIstart

            # Convert CGC to GDD mode
            crop.CGC_CD = crop.CGC
            crop.CGC = (np.log((((0.98*crop.CCx)-crop.CCx)*crop.CC0)
                                      /(-0.25*(crop.CCx**2)))
                              )/(-(crop.MaxCanopy-crop.Emergence))

            # Convert CDC to GDD mode
            crop.CDC_CD = crop.CDC
            tCD = crop.MaturityCD-crop.SenescenceCD
            if tCD <= 0:
                tCD = 1

            CCi = crop.CCx*(1-0.05*(np.exp((crop.CDC_CD/crop.CCx)*tCD)-1))
            if CCi < 0:
                CCi = 0

            tGDD = crop.Maturity-crop.Senescence
            if tGDD <= 0:
                tGDD = 5

            crop.CDC = (crop.CCx/tGDD)*np.log(1+((1-CCi/crop.CCx)/0.05))
            # Set calendar type to GDD mode
            crop.CalendarType = 2

    elif Mode ==2:
        # Growth in growing degree days
        # Time from sowing to end of vegatative growth period
        if crop.Determinant == 1:
            crop.CanopyDevEnd = round(crop.HIstart+(crop.Flowering/2))
        else:
            crop.CanopyDevEnd = crop.Senescence

        # Time from sowing to 10# canopy cover (non-stressed conditions)
        crop.Canopy10Pct = round(crop.Emergence+(np.log(0.1/crop.CC0)/crop.CGC))

        # Time from sowing to maximum canopy cover (non-stressed conditions)
        crop.MaxCanopy = round(crop.Emergence+(np.log((0.25*crop.CCx*crop.CCx/crop.CC0)
                                                                    /(crop.CCx-(0.98*crop.CCx)))/crop.CGC))

        # Time from sowing to end of yield formation
        crop.HIend = crop.HIstart+crop.YldForm

        # Time from sowing to end of flowering (if fruit/grain crop)
        if crop.CropType == 3:
            crop.FloweringEnd = crop.HIstart+crop.Flowering

        # Extract weather data for first growing season that crop is planted
        for i,n in enumerate(ParamStruct.CropChoices):
            if n == crop.Name:
                idx = i
                break
            else:
                idx = -1
        assert idx> -1
        date_range = pd.date_range(ClockStruct.PlantingDates[idx],ClockStruct.HarvestDates[idx])
        wdf = weather_df.copy(); wdf.index = wdf.Date

        wdf = wdf.loc[date_range]
        Tmin = wdf.MinTemp
        Tmax = wdf.MaxTemp


        # Calculate GDD's
        if crop.GDDmethod == 1:

            Tmean = (Tmax+Tmin)/2
            Tmean = Tmean.clip(lower=crop.Tbase,upper=crop.Tupp)
            GDD = Tmean-crop.Tbase

        elif crop.GDDmethod == 2:

            Tmax = Tmax.clip(lower=crop.Tbase,upper=crop.Tupp)
            Tmin = Tmin.clip(lower=crop.Tbase,upper=crop.Tupp)
            Tmean = (Tmax+Tmin)/2
            GDD = Tmean-crop.Tbase

        elif crop.GDDmethod == 3:

            Tmax = Tmax.clip(lower=crop.Tbase,upper=crop.Tupp)
            Tmin = Tmin.clip(upper=crop.Tupp)
            Tmean = (Tmax+Tmin)/2
            Tmean = Tmean.clip(lower=crop.Tbase)
            GDD = Tmean-crop.Tbase

        GDDcum = np.cumsum(GDD).reset_index(drop=True)

        # 1. GDD's from sowing to maximum canopy cover
        crop.MaxCanopyCD = (GDDcum>crop.MaxCanopy).idxmax()+1
        # 2. GDD's from sowing to end of vegetative growth
        crop.CanopyDevEndCD = (GDDcum>crop.CanopyDevEnd).idxmax()+1
        # 3. Calendar days from sowing to start of yield formation
        crop.HIstartCD = (GDDcum>crop.HIstart).idxmax()+1
        # 4. Calendar days from sowing to end of yield formation
        crop.HIendCD = (GDDcum>crop.HIend).idxmax()+1
        # 5. Duration of yield formation in calendar days
        crop.YldFormCD = crop.HIendCD-crop.HIstartCD
        if crop.CropType == 3:
            # 1. Calendar days from sowing to end of flowering
            FloweringEnd = (GDDcum>crop.FloweringEnd).idxmax()+1
            # 2. Duration of flowering in calendar days
            crop.FloweringCD = FloweringEnd-crop.HIstartCD


    return crop

# Cell
def calculate_HIGC(crop):
    """
    Function to calculate harvest index growth coefficient

    *Arguments:*\n

    `crop` : `CropClass` :  Crop object containing crop paramaters


    *Returns:*

    `crop` : `CropClass` : updated Crop object


    """
    # Determine HIGC
    # Total yield formation days
    tHI = crop.YldFormCD
    # Iteratively estimate HIGC
    HIGC = 0.001
    HIest = 0
    while HIest <= (0.98*crop.HI0):
        HIGC = HIGC+0.001
        HIest = (crop.HIini*crop.HI0)/(crop.HIini+(crop.HI0-crop.HIini)*np.exp(-HIGC*tHI))

    if HIest >= crop.HI0:
        HIGC = HIGC-0.001


    crop.HIGC = HIGC

    return crop

# Cell
def calculate_HI_linear(crop):

    """
    Function to calculate time to switch to linear harvest index build-up,
    and associated linear rate of build-up. Only for fruit/grain crops.

    *Arguments:*\n

    `crop` : `CropClass` :  Crop object containing crop paramaters


    *Returns:*

    `crop` : `CropClass` : updated Crop object


    """
    # Determine linear switch point
    # Initialise variables
    ti = 0
    tmax = crop.YldFormCD
    HIest = 0
    HIprev = crop.HIini
    # Iterate to find linear switch point
    while (HIest <= crop.HI0) and (ti < tmax):
        ti = ti+1
        HInew = (crop.HIini*crop.HI0)/(crop.HIini+(crop.HI0-crop.HIini)*np.exp(-crop.HIGC*ti))
        HIest = HInew+(tmax-ti)*(HInew-HIprev)
        HIprev = HInew

    tSwitch = ti-1

    # Determine linear build-up rate
    if tSwitch > 0:
        HIest = (crop.HIini*crop.HI0)/(crop.HIini+(crop.HI0-crop.HIini)*np.exp(-crop.HIGC*tSwitch))
    else:
        HIest = 0

    dHILin = (crop.HI0-HIest)/(tmax-tSwitch)


    crop.tLinSwitch = tSwitch
    crop.dHILinear = dHILin

    return crop

# Cell
def read_model_initial_conditions(ParamStruct,ClockStruct,InitWC):
    """
    Function to set up initial model conditions

    *Arguments:*\n

    `ParamStruct` : `ParamStructClass` :  Contains model crop and soil paramaters

    `ClockStruct` : `ClockStructClass` :  model time paramaters

    `FileLocations` : `FileLocationsClass`:  input File Locations


    *Returns:*

    `ParamStruct` : `ParamStructClass` :  updated ParamStruct object

    `InitCond` : `InitCondClass` :  containing initial model paramaters


    """

    ###################
    # creat initial condition class
    ###################

    InitCond = InitCondClass()

    #Aeration stress (compartment level)
    InitCond.AerDaysComp = np.zeros(len(ParamStruct.Soil.profile))

    if ClockStruct.SeasonCounter == -1:
        InitCond.Zroot = 0
        InitCond.CC0adj = 0

    elif ClockStruct.SeasonCounter == 0:
        InitCond.Zroot = ParamStruct.Seasonal_Crop_List[0].Zmin
        InitCond.CC0adj = ParamStruct.Seasonal_Crop_List[0].CC0


    ##################
    # save field management
    ##################

    # Initial surface storage between any soil bunds
    if ClockStruct.SeasonCounter == -1:
        # First day of simulation is in fallow period
        if (ParamStruct.FallowFieldMngt.Bunds == 'Y') and (float(ParamStruct.FallowFieldMngt.zBund) > 0.001):
            # Get initial storage between surface bunds
            InitCond.SurfaceStorage = float(ParamStruct.FallowFieldMngt.BundWater)
            if InitCond.SurfaceStorage > float(ParamStruct.FallowFieldMngt.zBund):
                InitCond.SurfaceStorage = float(ParamStruct.FallowFieldMngt.zBund)
        else:
            # No surface bunds
            InitCond.SurfaceStorage = 0

    elif ClockStruct.SeasonCounter == 0:
        # First day of simulation is in first growing season
        # Get relevant field management structure parameters
        crop = ParamStruct.Seasonal_Crop_List[ClockStruct.SeasonCounter]
        FieldMngtTmp = crop.FieldMngt
        if (FieldMngtTmp.Bunds == 'Y') and (float(FieldMngtTmp.zBund) > 0.001):
            # Get initial storage between surface bunds
            InitCond.SurfaceStorage = float(FieldMngtTmp.BundWater)
            if InitCond.SurfaceStorage > float(FieldMngtTmp.zBund):
                InitCond.SurfaceStorage = float(FieldMngtTmp.zBund)
        else:
            # No surface bunds
            InitCond.SurfaceStorage = 0


    ############
    # watertable
    ############



    profile = ParamStruct.Soil.profile


    # Check for presence of groundwater table
    if ParamStruct.WaterTable == 0: # No water table present
        # Set initial groundwater level to dummy value
        InitCond.zGW = -999
        InitCond.WTinSoil = False
        # Set adjusted field capacity to default field capacity
        InitCond.th_fc_Adj = profile.th_fc.values
    elif ParamStruct.WaterTable == 1: # Water table is present
        # Set initial groundwater level
        InitCond.zGW = ParamStruct.zGW.loc[ClockStruct.StepStartTime]
        # Find compartment mid-points
        zMid = profile.zMid
        # Check if water table is within modelled soil profile
        if InitCond.zGW >= 0:
            idx = zMid[zMid>=InitCond.zGW].index
            if idx.shape[0]==0:
                InitCond.WTinSoil = False
            else:
                InitCond.WTinSoil = True
        else:
            InitCond.WTinSoil = False



        # Adjust compartment field capacity
        compi = int(len(profile))-1
        thfcAdj = np.zeros(compi+1)
        while compi >= 0:
            # get soil layer of compartment
            compdf = profile.loc[compi]
            if compdf.th_fc <= 0.1:
                Xmax = 1
            else:
                if compdf.th_fc >= 0.3:
                    Xmax = 2
                else:
                    pF = 2+0.3*(compdf.th_fc-0.1)/0.2
                    Xmax = (np.exp(pF*np.log(10)))/100

            if (InitCond.zGW < 0) or ((InitCond.zGW-zMid.iloc[compi]) >= Xmax):
                for ii in range(compi):
                    compdfii = profile.loc[ii]
                    thfcAdj[ii] = compdfii.th_fc

                compi = -1
            else:
                if compdf.th_fc >= compdf.th_s:
                    thfcAdj[compi] = compdf.th_fc
                else:
                    if zMid.iloc[compi] >= InitCond.zGW:
                        thfcAdj[compi] = compdf.th_s
                    else:
                        dV = compdf.th_s-compdf.th_fc
                        dFC = (dV/(Xmax**2))*((zMid.iloc[compi]-(InitCond.zGW-Xmax))**2)
                        thfcAdj[compi] = compdf.th_fc+dFC


                compi = compi-1


        # Store adjusted field capacity values
        InitCond.th_fc_Adj = np.round(thfcAdj,3)

    profile["th_fc_Adj"] = np.round(InitCond.th_fc_Adj,3)



    # create hydrology df to group by layer instead of compartment
    ParamStruct.Soil.Hydrology = profile.groupby('Layer').mean().drop(['dz','dzsum'],axis=1)
    ParamStruct.Soil.Hydrology["dz"] = profile.groupby('Layer').sum().dz


    ###################
    # initial water contents
    ###################



    typestr = InitWC.wc_type
    methodstr = InitWC.Method

    depth_layer = InitWC.depth_layer
    datapoints = InitWC.value

    values=np.zeros(len(datapoints))

    hydf = ParamStruct.Soil.Hydrology

    # Assign data
    if typestr == 'Num':
        # Values are defined as numbers (m3/m3) so no calculation required
        depth_layer = np.array(depth_layer,dtype=float)
        values = np.array(datapoints,dtype=float)

    elif typestr == 'Pct':
        # Values are defined as percentage of TAW. Extract and assign value for
        # each soil layer based on calculated/input soil hydraulic properties
        depth_layer = np.array(depth_layer,dtype=float)
        datapoints = np.array(datapoints,dtype=float)

        for ii in range(len(values)):
            if methodstr == 'Depth':
                depth = depth_layer[ii]
                value = datapoints[ii]


                # Find layer at specified depth
                if depth < profile.dzsum.iloc[-1]:
                    layer = profile.query(f'{depth}<dzsum').Layer.iloc[0]
                else:
                    layer = profile.Layer.iloc[-1]

                compdf = hydf.loc[layer]

                # Calculate moisture content at specified depth
                values[ii] = compdf.th_wp+((value/100)*(compdf.th_fc-compdf.th_wp))
            elif methodstr == 'Layer':
                # Calculate moisture content at specified layer
                layer = depth_layer[ii]
                value = datapoints[ii]

                compdf = hydf.loc[layer]

                values[ii] = compdf.th_wp+((value/100)*(compdf.th_fc-compdf.th_wp))

    elif typestr=='Prop':
        # Values are specified as soil hydraulic properties (SAT, FC, or WP).
        # Extract and assign value for each soil layer

        for ii in range(len(values)):
            if methodstr=='Depth':
                # Find layer at specified depth
                depth = depth_layer[ii]
                value = datapoints[ii]

                # Find layer at specified depth
                if depth < profile.dzsum.iloc[-1]:
                    layer = profile.query(f'{depth}<dzsum').Layer.iloc[0]
                else:
                    layer = profile.Layer.iloc[-1]

                compdf = hydf.loc[layer]

                # Calculate moisture content at specified depth
                if value=='SAT':
                    values[ii] = compdf.th_s
                if value=='FC':
                    values[ii] = compdf.th_fc
                if value=='WP':
                    values[ii] = compdf.th_wp

            elif methodstr=='Layer':
                # Calculate moisture content at specified layer
                layer = depth_layer[ii]
                value = datapoints[ii]

                compdf = hydf.loc[layer]

                if value=='SAT':
                    values[ii] = compdf.th_s
                if value=='FC':
                    values[ii] = compdf.th_fc
                if value=='WP':
                    values[ii] = compdf.th_wp


    # Interpolate values to all soil compartments

    thini = np.zeros(int(profile.shape[0]))
    if methodstr=='Layer':
        for ii in range(len(values)):
            layer = depth_layer[ii]
            value = values[ii]

            idx =  profile.query(f'Layer=={int(layer)}').index

            thini[idx] = value

        InitCond.th = thini

    elif methodstr=='Depth':
        depths = depth_layer

        # Add zero point
        if depths[0] > 0:
            depths = np.append([0],depths)
            values = np.append([values[0]],values)

        # Add end point (bottom of soil profile)
        if depths[-1] < ParamStruct.Soil.zSoil:
            depths = np.append(depths,[ParamStruct.Soil.zSoil])
            values = np.append(values,[values[-1]])

        # Find centroids of compartments
        SoilDepths = profile.dzsum.values
        comp_top = np.append([0],SoilDepths[:-1])
        comp_bot = SoilDepths
        comp_mid = (comp_top+comp_bot)/2
        # Interpolate initial water contents to each compartment
        thini = np.interp(comp_mid,depths,values)
        InitCond.th = thini

    # If groundwater table is present and calculating water contents based on
    # field capacity, then reset value to account for possible changes in field
    # capacity caused by capillary rise effects
    if ParamStruct.WaterTable == 1:
        if (typestr=='Prop') and (datapoints[-1]=='FC'):
            InitCond.th = InitCond.th_fc_Adj




    # If groundwater table is present in soil profile then set all water
    # contents below the water table to saturation
    if InitCond.WTinSoil == True:
        # Find compartment mid-points
        SoilDepths = profile.dzsum.values
        comp_top = np.append([0],SoilDepths[:-1])
        comp_bot = SoilDepths
        comp_mid = (comp_top+comp_bot)/2
        idx = comp_mid[comp_mid>=InitCond.zGW].iloc[0]
        for ii in range(idx,len(profile)):
            layeri = profile.loc[ii].Layer
            InitCond.th[ii] = hydf.th_s.loc[layeri]

    InitCond.thini = InitCond.th

    ParamStruct.Soil.profile = profile
    ParamStruct.Soil.Hydrology = hydf



    return ParamStruct, InitCond

# Cell
def create_soil_profile(ParamStruct):
    """
    funciton to create soil profile class to store soil info. Its much faster to access
    the info when its in a class compared to a dataframe

    *Arguments:*\n

    `ParamStruct` : `ParamStructClass` :  Contains model crop and soil paramaters

    *Returns:*

    `ParamStruct` : `ParamStructClass` :  updated with soil profile


    """

    Profile = SoilProfileClass()

    for c in range(ParamStruct.Soil.profile.shape[0]):
        comdf = ParamStruct.Soil.profile.iloc[c]
        Comp = SoilCompClass()
        Comp.Comp = comdf.Comp.item()
        Comp.dz = comdf.dz.item()
        Comp.Layer = comdf.Layer.item()
        Comp.dzsum = comdf.dzsum.item()
        Comp.th_fc = comdf.th_fc.item()
        #Comp.Layer_dz = comdf.Layer_dz.item()
        Comp.th_s = comdf.th_s.item()
        Comp.th_wp = comdf.th_wp.item()
        Comp.Ksat = comdf.Ksat.item()
        Comp.Penetrability = comdf.penetrability.item()
        Comp.th_dry = comdf.th_dry.item()
        Comp.tau = comdf.tau.item()
        Comp.zBot = comdf.zBot.item()
        Comp.zTop = comdf.zTop.item()
        Comp.zMid = comdf.zMid.item()
        Comp.th_fc_Adj = comdf.th_fc_Adj.item()
        if ParamStruct.WaterTable==1:
            Comp.aCR = comdf.aCR.item()
            Comp.bCR = comdf.bCR.item()

        Profile.Comp.append(Comp)

    Profile.dz = ParamStruct.Soil.profile.dz.values
    Profile.dzsum = ParamStruct.Soil.profile.dzsum.values
    Profile.zBot = ParamStruct.Soil.profile.zBot.values
    Profile.zTop = ParamStruct.Soil.profile.zTop.values
    Profile.zMid = ParamStruct.Soil.profile.zMid.values


    # add empty layer as layer list starts at 1
    Profile.Layer.append(SoilLayerClass())

    for c in range(ParamStruct.Soil.Hydrology.shape[0]):
        ldf = ParamStruct.Soil.Hydrology.iloc[c]
        Layer = SoilLayerClass()
        Layer.dz = ldf.dz.item()
        Layer.th_fc = ldf.th_fc.item()
        Layer.th_s = ldf.th_s.item()
        Layer.th_wp = ldf.th_wp.item()
        Layer.Ksat = ldf.Ksat.item()
        Layer.Penetrability = ldf.penetrability.item()
        Layer.th_dry = ldf.th_dry.item()
        Layer.tau = ldf.tau.item()
        if ParamStruct.WaterTable==1:
            Layer.aCR = ldf.aCR.item()
            Layer.bCR = ldf.bCR.item()

        Profile.Layer.append(Layer)


    ParamStruct.Soil.Profile = Profile


    return ParamStruct
